/*===================================================================================
**
**	Author	:	Robert Bateman
**	E-Mail	:	rbateman@bournemouth.ac.uk
**
**	Copyright (c) 2002
**
**	National Centre for Computer Animation,
**	Bournemouth University,
**	Talbot Campus,
**	Bournemouth,
**	BH3 72F,
**	England
**
**	Not for commercial use.
**
===================================================================================*/

/*===================================================================================
**
**		Includes
**
**=================================================================================*/

#include <stdio.h>
#include <GL/glut.h>
#include "FileDialog.h"

/*===================================================================================
**
**		A nasty hack so that i dont have to include ObjLoader.h
**		This is needed because there are 6 versions of the file in this tutorial. 
**		
**
**=================================================================================*/

typedef unsigned int ObjFile;

extern ObjFile	LoadOBJ(const char*);
extern void		DrawOBJ(ObjFile);
extern void		DeleteOBJ(ObjFile);
extern void		CleanUpOBJ(void);
extern void		ConvertMeshTo(ObjFile,unsigned char);
extern void		SetLightPosition(float lx,float ly,float lz);
extern void		SetTextures(ObjFile id,const char BumpTex[],const char BaseTex[]);

#define DISPLAY_LIST 1
#define VERTEX_ARRAY 2
#define VERTEX_BUFFER_OBJECT 3

/*===================================================================================
**
**		Global Variables for the File Dialog (gtk) under linux
**
**=================================================================================*/

#ifndef WIN32
extern int g_argc;
extern char** g_argv;
#endif

/*===================================================================================
**
**		Maya Style Mouse Movement Global Variables
**
**=================================================================================*/

unsigned int g_bLMB=0,
			 g_bMMB=0;

int g_iLastPosX,
	g_iLastPosY,
	g_iStepX   ,
	g_iStepY   ;

float g_fZoom   = 0.0f,
	  g_fTransX = 0.0f,
	  g_fTransY = 0.0f,
	  g_fSpinX  = 0.001f,
	  g_fSpinY  = 0.001f;


float lpos[4] = {0,10,0,1};

/*===================================================================================
**
**		Function Prototypes
**
**=================================================================================*/

void DoTransform(void);
void InitGL(void);
void Display(void);
void ButtonPress(int button,int state,int x,int y);
void MouseMotion(int x,int y);
void Reshape(int w,int h);


/*===================================================================================
**
**		Stuff We are worried about
**
**=================================================================================*/

/*
**	References to the files
*/
ObjFile Objects[20];
int NumObjects=0;


/*
**	GLUT display callback function
*/
void Display(void)
{
	int i;

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	DoTransform();

	glLightfv(GL_LIGHT0,GL_POSITION,lpos);

	/*
	**	Draw the model
	*/
	for(i=0;i!=NumObjects;++i)
	{
		DrawOBJ(Objects[i]);
	}

	glPushMatrix();

		glTranslatef(lpos[0],lpos[1],lpos[2]);
		glutSolidSphere(0.2f,10,10);

	glPopMatrix();

	glutSwapBuffers();
}

/*
**	GLUT Callback function to deal with mouse movement
*/
void Keyboard(unsigned char Key,int x,int y)
{
	if (Key=='L' || Key=='l') {
		if (NumObjects<20) {

			/* used to store the returned filename from the file dialog */
			char filename[256];

			/* get a file dialog to find the file to load */
			if ( OpenFileDialog(filename) ) 
			{
				/* attempt to load the file */
				Objects[ NumObjects ] = LoadOBJ(filename);

				/* if file loaded ok */
				if (Objects[ NumObjects ] != 0) {
					ConvertMeshTo( Objects[ NumObjects ], VERTEX_BUFFER_OBJECT );
					SetTextures(Objects[ NumObjects ],"bump.tga","base.tga");
					++NumObjects;

				}
			}
		}
	}

	if (Key == 'w') 
		lpos[2] += 0.2f;

	if (Key == 's') 
		lpos[2] -= 0.2f;

	if (Key == 'd') 
		lpos[0] += 0.2f;

	if (Key == 'a') 
		lpos[0] -= 0.2f;

	if (Key == 'q') 
		lpos[1] += 0.2f;

	if (Key == 'e') 
		lpos[1] -= 0.2f;

	SetLightPosition(lpos[0],lpos[1],lpos[2]);

	glutPostRedisplay();
}

/*============================== As simple as that! ================================*/


/*
**	GLUT Callback function to deal with button presses and releases of the mouse
*/
void ButtonPress(int button,int state,int x,int y)
{
    if(state == GLUT_DOWN)
	{
        g_iLastPosX = x;
        g_iLastPosY = y;
        switch(button)
		{
		case GLUT_LEFT_BUTTON:
			g_bLMB = 1;
            break;
		case GLUT_MIDDLE_BUTTON:
		case GLUT_RIGHT_BUTTON:
			g_bMMB = 1;
            break;
		default:
            break;
		}
	}
	else
	{
		switch(button)
		{
		case GLUT_LEFT_BUTTON:
			g_bLMB = 0;
			break;
		case GLUT_MIDDLE_BUTTON:
		case GLUT_RIGHT_BUTTON:
			g_bMMB = 0;
            break;
		default:
			break;
		}
	}
	glutPostRedisplay();
}

/*
**	GLUT Callback function to deal with mouse movement
*/
void MouseMotion(int x,int y)
{
    g_iStepX = x-g_iLastPosX;
    g_iStepY = y-g_iLastPosY;

	if(g_bLMB && g_bMMB)
	{
		g_fZoom += 0.2f*g_iStepX;
	}
	else
	if(g_bLMB)
	{
		g_fSpinX += 0.5f*g_iStepY;
		g_fSpinY += 0.5f*g_iStepX;
	}
	else
	if(g_bMMB)
	{
		g_fTransX  += 0.2f*g_iStepX;
		g_fTransY  -= 0.2f*g_iStepY;
	}
    g_iLastPosX = x;
    g_iLastPosY = y;
	glutPostRedisplay();
}



/*
**	Function to impliment the transform required for the Maya Mouse movement
*/
void DoTransform(void)
{
	glLoadIdentity();
    gluLookAt(0,1,50,0,0,0,0,1,0);
    glTranslatef(g_fTransX,g_fTransY,g_fZoom);
    glRotatef   (g_fSpinX,1.0f,0.0f,0.0f);
    glRotatef   (g_fSpinY,0.0f,1.0f,0.0f);
}


/*
**	OpenGL Initialisation Function
*/
void InitGL(void)
{
	glClearColor(0.5f,0.5f,0.5f,0.0f);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	/*
	 *	Setup Material Porperties & colours for the light
	 */
	{
		float Ambient[] = {0.4f,0.4f,0.4f,0.0f};
		float Diffuse[] = {0.75f,0.75f,0.75f,0.0f};

		glMaterialfv(GL_FRONT,GL_AMBIENT,Ambient);
		glMaterialfv(GL_FRONT,GL_DIFFUSE,Diffuse);
	}
	{
		float Ambient[] = {0.4f,0.4f,0.4f,0.0f};
		float Diffuse[] = {0.95f,0.95f,0.95f,0.0f};

		glLightfv(GL_LIGHT0,GL_AMBIENT,Ambient);
		glLightfv(GL_LIGHT0,GL_DIFFUSE,Diffuse);
	}
}

/*
**	GLUT callback function to deal with resizing of the window
*/
void Reshape( int w , int h )
{
	/* prevent a crash when minimising */
	if(h==0) h=1;
	glViewport(0, (GLsizei) 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f,(float)(w/((h==0)?1:h)),0.01f,500.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

const char* Instructions=	"ObjLoader Examples\n"
							"------------------\n"
							"\n"
							"LMB     - Rotate\n"
							"MMB     - Track\n"
							"RMB     - Track\n"
							"LMB+MMB - Zoom\n"
							"LMB+RMB - Zoom\n"
							"\n"
							"L       - Load Obj File\n";

int main( int argc , char **argv )
{
#ifndef WIN32
	g_argc=argc;
	g_argv=argv;
#endif

	printf(Instructions);

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

	glutInitWindowSize(640,480);
	glutInitWindowPosition(100,100);
	glutCreateWindow("Obj Loader");


	/* register program callbacks with glut	*/

	glutDisplayFunc ( Display );
	glutReshapeFunc ( Reshape );
	glutMouseFunc   ( ButtonPress );
	glutMotionFunc  ( MouseMotion );
	glutKeyboardFunc( Keyboard );

	/* do initialisation */

	InitGL();

	/* enter main loop */
	glutMainLoop();
	return 0;
}
